#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
#
# DO NOT ALTER OR REMOVE COPYRIGHT NOTICES OR THIS HEADER.
#
# Copyright 2008 Sun Microsystems, Inc. All rights reserved.
#
# The contents of this file are subject to the terms of either the GNU
# General Public License Version 2 only ("GPL") or the Common Development
# and Distribution License("CDDL") (collectively, the "License").  You
# may not use this file except in compliance with the License. You can obtain
# a copy of the License at https://glassfish.dev.java.net/public/CDDL+GPL.html
# or updatetool/LICENSE.txt.  See the License for the specific
# language governing permissions and limitations under the License.
#
# When distributing the software, include this License Header Notice in each
# file and include the License file at glassfish/bootstrap/legal/LICENSE.txt.
# Sun designates this particular file as subject to the "Classpath" exception
# as provided by Sun in the GPL Version 2 section of the License file that
# accompanied this code.  If applicable, add the following below the License
# Header, with the fields enclosed by brackets [] replaced by your own
# identifying information: "Portions Copyrighted [year]
# [name of copyright owner]"
#
# Contributor(s):
#
# If you wish your version of this file to be governed by only the CDDL or
# only the GPL Version 2, indicate your decision by adding "[Contributor]
# elects to include this software in this distribution under the [CDDL or GPL
# Version 2] license."  If you don't indicate a single choice of license, a
# recipient has the option to distribute your version of this file under
# either the CDDL, the GPL Version 2 or to extend the choice of license to
# its licensees as provided above.  However, if you add GPL Version 2 code
# and therefore, elected the GPL Version 2 license, then the option applies
# only if the new code is made subject to such option by the copyright
# holder.
#
import errno
import os
import sys
import threading
import time
import socket
import urllib2
import httplib
import pkg.bundle
import pkg.depot as depot
import pkg.server.face as face
import pkg.server.config as config
import pkg.server.repository as repo
from pkg.misc import port_available, versioned_urlopen

import cherrypy

class DepotStateException(Exception):

        def __init__(self, reason):
                Exception.__init__(self, reason)
         
class DepotController(threading.Thread):

        HALTED = 0
        STARTING = 1
        RUNNING = 2

        def __init__(self, port=10001, repo='.'):
                threading.Thread.__init__(self)
                self.__port = port
                self.__dir = repo
                self.__state = DepotController.HALTED
                self.setDaemon(True)
                return

        def get_depot_url(self):
                return "http://localhost:%d" % self.__port

        def __network_ping(self):
                try:
                    c, v = versioned_urlopen(self.get_depot_url(), "catalog", [0])
                except urllib2.HTTPError, e:
                    # Server returns NOT_MODIFIED if catalog is up
                    # to date
                    if e.code == httplib.NOT_MODIFIED:
                        return True
                    else:
                        return False
                except urllib2.URLError:
                        return False
                return True

        def is_alive(self):
                """ First, check that the depot process seems to be alive.
                    Then make a little HTTP request to see if the depot is
                    responsive to requests """
                if not self.isAlive():
                    return False
                return self.__network_ping()


        def get_args(self):
                """ Return the equivalent command line invocation (as an
                    array) for the depot as currently configured. """

                args =  [  ]
                if self.__port != -1:
                        args.append("-p")
                        args.append("%d" % self.__port)
                if self.__dir != None:
                        args.append("-d")
                        args.append(self.__dir)
                return args
                
        def run(self):
                if not port_available("localhost", self.__port):
                        raise DepotStateException("A depot (or some " +
                                    "other network process) seems to be " +
                                    "running on port %d already!" % self.__port)

                self.__state = DepotController.STARTING
                depot.scfg = config.SvrConfig(self.__dir, depot.AUTH_DEFAULT)
                depot.scfg.init_dirs()
                depot.scfg.acquire_in_flight()
                depot.scfg.acquire_catalog()
                root = cherrypy.Application(repo.Repository(depot.scfg))
                # We have to override cherrypy's default response_class so that we
                # have access to the write() callable to stream data directly to the
                # client.
                root.wsgiapp.response_class = pkg.server.depot.DepotResponse

                cherrypy.config.update({
                    "environment": "production",
                    "checker.on": True,
                    "log.screen": False,
                    "server.socket_port": self.__port,
                    "server.thread_pool": depot.THREADS_DEFAULT,
                    "server.socket_timeout": depot.SOCKET_TIMEOUT_DEFAULT
                })

                conf = {
                    "/robots.txt": {
                        "tools.staticfile.on": True,
                        "tools.staticfile.filename": os.path.join(face.content_root,
                            "robots.txt")
                    },
                    "/static": {
                        "tools.staticdir.on": True,
                        "tools.staticdir.root": face.content_root,
                        "tools.staticdir.dir": ""
                    }
                }

                try:
                        cherrypy.quickstart(root, config = conf)
                except:
                        print "pkg.depotd: unknown error starting depot, illegal " \
                            "option value specified?"
                        usage()
            
        def waitforup(self):
                sleeptime = 0.05
                contact = False
                while sleeptime <= 20.0:
                        if self.is_alive():
                                contact = True
                                break
                        time.sleep(sleeptime)
                        sleeptime *= 2
                      
                if contact == False:
                        self.__state = DepotController.HALTED
                        raise DepotStateException("Depot did not respond to repeated attempts to make contact")

                self.__state = DepotController.RUNNING
