#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# DO NOT ALTER OR REMOVE COPYRIGHT NOTICES OR THIS HEADER.
#
# Copyright 2008 Sun Microsystems, Inc. All rights reserved.
#
# The contents of this file are subject to the terms of either the GNU
# General Public License Version 2 only ("GPL") or the Common Development
# and Distribution License("CDDL") (collectively, the "License").  You
# may not use this file except in compliance with the License. You can obtain
# a copy of the License at https://glassfish.dev.java.net/public/CDDL+GPL.html
# or updatetool/LICENSE.txt.  See the License for the specific
# language governing permissions and limitations under the License.
#
# When distributing the software, include this License Header Notice in each
# file and include the License file at glassfish/bootstrap/legal/LICENSE.txt.
# Sun designates this particular file as subject to the "Classpath" exception
# as provided by Sun in the GPL Version 2 section of the License file that
# accompanied this code.  If applicable, add the following below the License
# Header, with the fields enclosed by brackets [] replaced by your own
# identifying information: "Portions Copyrighted [year]
# [name of copyright owner]"
#
# Contributor(s):
#
# If you wish your version of this file to be governed by only the CDDL or
# only the GPL Version 2, indicate your decision by adding "[Contributor]
# elects to include this software in this distribution under the [CDDL or GPL
# Version 2] license."  If you don't indicate a single choice of license, a
# recipient has the option to distribute your version of this file under
# either the CDDL, the GPL Version 2 or to extend the choice of license to
# its licensees as provided above.  However, if you add GPL Version 2 code
# and therefore, elected the GPL Version 2 license, then the option applies
# only if the new code is made subject to such option by the copyright
# holder.
#

# makepkgs - script for creating Update Center IPS packages

import builder
import errno
import os
import sys
import imp
import tempfile
import time
import traceback
import getopt
import urllib2
import httplib
import depotcontroller
import pkg.bundle
import pkg.publish.transaction as trans
import pkg.config as config
import pkg.depot as depot
import tarfile
from pkg.misc import versioned_urlopen
from tarpkgs import makeTar

seenfiles = []

def addfile(f, pcfg, id, prefix, attributes, opsys):

    # defaults
    mode = "0644"
    owner = "nobody"
    group = "sys"

    if f in seenfiles:
        # skip this file because we've already inserted it
        return
    seenfiles.append(f)

    if attributes and "file" in attributes:
        pathname = attributes["file"]
    else:
        pathname = os.path.join(prefix, f)

    # If file is a symlink then use symlink action
    if os.path.islink(pathname):
        addlink(f, pcfg, id, prefix)
        return

    # If a file is exeuctable, default to execute permission
    if os.access(pathname, os.X_OK):
        mode = "0755"

    if (attributes != None):
        if attributes.has_key("mode"): mode = attributes["mode"]
        if attributes.has_key("owner"): mode = attributes["owner"]
        if attributes.has_key("group"): mode = attributes["group"]
        if attributes.has_key("os") and attributes["os"] != opsys:
            # skip this file because it is not for this os
            return

    l = ['mode=' + mode, 'owner=' + owner, 'group=' + group, 'path=' + f]
    action = pkg.actions.fromlist("file", l)
    action.data = lambda : open(pathname, "rb")
    t = trans.Transaction()
    status, msg, body = t.add(pcfg, id, action)
    print "add of file %s %s: status=%d, msg=%s, body=%s" % (f, l, status, msg, body)
    if status != 200:
        print >> sys.stderr, "unsuccessful status %d when adding file" % status
        sys.exit(1)

def adddir(d, pcfg, id, prefix, attributes, opsys):

    # defaults
    mode = "0755"
    owner = "nobody"
    group = "sys"

    if (attributes != None):
        if attributes.has_key("mode"): mode = attributes["mode"]
        if attributes.has_key("owner"): mode = attributes["owner"]
        if attributes.has_key("group"): mode = attributes["group"]
        if attributes.has_key("os") and attributes["os"] != opsys:
            # skip this file because it is not for this os
            return

    action = pkg.actions.fromlist("dir", ['mode=' + mode, 'owner=' + owner, 'group=' + group, 'path=' + d])
    t = trans.Transaction()
    status, msg, body = t.add(pcfg, id, action)
    print "add of directory %s: status=%d, msg=%s, body=%s" % (d, status, msg, body)
    if status != 200:
        print >> sys.stderr, "unsuccessful status %d when adding directory" % status
        sys.exit(1)

def addlink(path, pcfg, id, prefix):

    target_path = os.readlink(os.path.join(prefix, path))

    action = pkg.actions.fromlist("link", ['path=' + path, \
                                           'target=' + target_path])
    t = trans.Transaction()
    status, msg, body = t.add(pcfg, id, action)
    print "add of symlink %s->%s: status=%d, msg=%s, body=%s" % (path, target_path, status, msg, body)
    if status != 200:
        print >> sys.stderr, "unsuccessful status %d when adding link" % status
        sys.exit(1)

def addlicense(f, pcfg, id, prefix, attributes, opsys):

    pathname = os.path.join(prefix, f)
    lic_name = attributes["license"]

    if attributes.has_key("os") and attributes["os"] != opsys:
        # skip this file because it is not for this os
        return

    l = ['license=' + lic_name]
    action = pkg.actions.fromlist("license", l)
    action.data = lambda : open(os.path.join(prefix, f), "rb")
    t = trans.Transaction()
    status, msg, body = t.add(pcfg, id, action)
    print "add of license %s %s: status=%d, msg=%s, body=%s" % (f, l, status, msg, body)
    if status != 200:
        print >> sys.stderr, "unsuccessful status %d when adding license" % status
        sys.exit(1)

def ping_repo(url):
    try:
        c, v = versioned_urlopen(url, "catalog", [0])
    except urllib2.HTTPError, e:
        # Server returns NOT_MODIFIED if catalog is up
        # to date
        if e.code == httplib.NOT_MODIFIED:
            return True
        else:
            print >> sys.stderr, "unable to access repository catalog:", e
            return False
    return True

def usage(msg):
    ret = """Usage: %s {-d repodir | -s repourl} [-o sys] -b basedir
Options:
    -d repodir   repository directory to publish to
    -s repourl   repository URL to publish to
    -t tgzfile   a .tgz file name to put a repository fragment into
    -b basedir   base direcory containing content
    -o sys       operating system identifier
    -v version   a version identifier to append to package version
    -D key=value pass property value to scripts. Accessible as makepkgs.props['key']
    """ % (__file__,)
    if msg is not None:
        ret = "%s\n\n%s" % (msg, ret)

    print >> sys.stderr, ret

def main_func():
    try:
        opts, pargs = getopt.getopt(sys.argv[1:], "d:b:s:t:o:v:D:")
    except getopt.GetoptError, e:
        usage("makepkgs: illegal option -- %s" % e.opt)
        return 1

    repo = None
    depoturl = None
    tgzfile = None
    base = "."
    opsys = ""
    version = ""
    builder.props = dict()
    for opt, arg in opts:
        if opt == "-d":
            repo = arg
        if opt == "-b":
            base = arg
        if opt == "-s":
            depoturl = arg
        if opt == "-t":
            tgzfile = arg;
        if opt == "-o":
            opsys = arg
        if opt == "-v":
            version = arg
        if opt == "-D":
            idx = arg.find('=')
            if idx<0:
                builder.props[arg]='true'
            else:
                builder.props[arg[:idx]]=arg[idx+1:]

    if (repo != None and depoturl != None) or (repo == None and depoturl == None):
        usage("invalid arguments: exactly one of -d or -s must be specified")
        return 1

    if len(pargs) == 0:
        usage("invalid arguments: no package modules listed")

    seendirs = []
    pkgs = []
    pkgnames = []
    for pm in pargs:
        if os.path.exists(pm):
            module_name = pm.replace('.', '_')
            module = imp.load_source(module_name, pm)
        else:
            module = __import__(pm)

        pkgs.append(module.pkg)
        pkgnames.append(module.pkg['name'])

    print "base directory:", base
    print "package list:", pkgnames

    if repo != None:
        dc = depotcontroller.DepotController(repo=repo)
        print "args:", dc.get_args()
        print "starting depot..."
        try:
            dc.start()
            dc.waitforup()
        except depotcontroller.DepotStateException, e:
            print "unable to start depot."
            sys.exit(1)
        print "depot started."
        depoturl = dc.get_depot_url()

    # check to see if the server is up
    if depoturl != None and not ping_repo(depoturl):
        sys.exit(1)
        
    pcfg = config.ParentRepo(depoturl, [depoturl])

    for p in pkgs:
        t = trans.Transaction()
        pname = "%s@%s%s" % (p["name"], p["version"], version)
        status, id = t.open(pcfg, pname)
        print "open of %s: status=%d, id=%s" % (pname, status, id)

        for an, av in p["attributes"].iteritems():
            al = ['name=' + an, 'value=' + av]
            action = pkg.actions.fromlist("set", al)
            t = trans.Transaction()
            status, msg, body = t.add(pcfg, id, action)
            print "add of %s: status=%d, msg=%s, body=%s" % (al, status, msg, body)
            if status != 200:
                print >> sys.stderr, "unsuccessful status %d when adding set" % status
                sys.exit(1)

        if p.has_key("depends"):
            for fmri, attributes in p["depends"].iteritems():
                dl = ['type=' + attributes["type"], 'fmri=' + fmri]
                action = pkg.actions.fromlist("depend", dl)
                t = trans.Transaction();
                status, msg, body = t.add(pcfg, id, action)
                print "add of depend %s: status=%d, msg=%s, body=%s" % (dl, status, msg, body)
                if status != 200:
                    print >> sys.stderr, "unsuccessful status %d when adding depend" % status
                    sys.exit(1)

        if p.has_key("files"):
            for path, attributes in p["files"].iteritems():
                addfile(path, pcfg, id, base, attributes, opsys)

        if p.has_key("dirs"):
            for path, attributes in p["dirs"].iteritems():
                adddir(path, pcfg, id, base, attributes, opsys)
                seendirs.append(path)

        if p.has_key("licenses"):
            for path, attributes in p["licenses"].iteritems():
                addlicense(path, pcfg, id, base, attributes, opsys)


        if p.has_key("dirtrees"):
            for d in p["dirtrees"]:
                if d not in seendirs:
                    adddir(d, pcfg, id, base, None, opsys)
                    seendirs.append(d)
                for root, dirs, files in os.walk(os.path.join(base, d)):
                    reldir = root[len(base)+1:]
                    for name in files:
                        f = os.path.join(reldir, name)
                        addfile(f, pcfg, id, base, None, opsys)
                    for name in dirs:
                        f = os.path.join(reldir, name)
                        # Only add directory if we haven't added it before
                        if f not in seendirs:
                            adddir(f, pcfg, id, base, None, opsys)
                            seendirs.append(f)

        t = trans.Transaction()
        ret, hdrs = t.close(pcfg, id, False)
        print "close: ret=%d, hdrs=%s" % (ret, hdrs)

    if repo != None:
        # ping the depot to make sure it has completed its last transaction
        dc.waitforup()
        print "depot stops when process exits."

    if tgzfile != None:
        if repo == None:
            print >> sys.stderr, "-t option requires -d"
            sys.exit(1)
        makeTar(repo,tarfile.open(tgzfile,"w:gz"),pkgnames)

    return 0

if __name__ == "__main__":
    try:
        ret = main_func()
    except SystemExit, e:
        raise e
    except KeyboardInterrupt:
        print "Interrupted"
        sys.exit(1)
    except:
        traceback.print_exc()
        sys.exit(99)
    sys.exit(ret)

