#!/usr/bin/python
#
# CDDL HEADER START
#
# The contents of this file are subject to the terms of the
# Common Development and Distribution License (the "License").
# You may not use this file except in compliance with the License.
#
# You can obtain a copy of the license at usr/src/OPENSOLARIS.LICENSE
# or http://www.opensolaris.org/os/licensing.
# See the License for the specific language governing permissions
# and limitations under the License.
#
# When distributing Covered Code, include this CDDL HEADER in each
# file and include the License file at usr/src/OPENSOLARIS.LICENSE.
# If applicable, add the following below this CDDL HEADER, with the
# fields enclosed by brackets "[]" replaced with your own identifying
# information: Portions Copyright [yyyy] [name of copyright owner]
#
# CDDL HEADER END
#
# Copyright 2008 Sun Microsystems, Inc.  All rights reserved.
# Use is subject to license terms.
#

"""
Windows has a specific implementation for most of these.  For the
group and user id-related (credential) APIs, no implementation
is provided.  This causes the file and directory actions to not
utilize any credential metadata when acting on Windows-compatible
systems. In the future, this may be able to be mapped onto the
NTFS group mechanism if deemed useful.
"""

import getpass
import shutil
import os
import errno
import tempfile
import util as os_util

def get_isainfo():
        """ TODO: Detect Windows 64-bit"""
        return ['i386']

def get_release():
        return os_util.get_os_release()

def get_platform():
        """ TODO: any other windows platforms to support?"""
        return 'i86pc'

def get_group_by_name(name, dirpath, use_file):
        """group names/numbers are ignored on Windows."""
        return -1

def get_user_by_name(name, dirpath, use_file):
        """group names/numbers are ignored on Windows."""
        return -1

def get_name_by_gid(gid, dirpath, use_file):
        """group names/numbers are ignored on Windows."""
        return ''

def get_name_by_uid(uid, dirpath, use_file):
        """group names/numbers are ignored on Windows."""
        return '' 

def get_username():
        try:
                return getpass.getuser()
        except ImportError:
                # getpass.getuser() will fail on systems without
                # pwd module, so try a common python windows add-on
                try:
                        import win32api
                        return win32api.GetUserName()
                except ImportError:
                        return None

def is_admin():
        try:
                # ctypes only available in python 2.5 or later
                import ctypes
                return ctypes.windll.shell32.IsUserAnAdmin() != 0
        except ImportError:
                return False

def chown(path, owner, group):
        """
        group names/numbers are ignored on Windows, so changing
        ownership of a file makes no sense.
        """
        return

# On Windows, rename to existing file is not allowed, so the
# destination file must be deleted first. But if the destination file 
# is open, it cannot be deleted. If the lock on the open file is 
# for a running executable, then the destination file can be renamed
# and deleted later when the file is no longer in use.
#
# To accomplish the delayed delete, the file is renamed to a trash folder
# within the image containing the file. A single image is assumed to be 
# contained within a single file system, thus making the rename feasible. 
# The cleanup_trash method is called at the end of rename to cleanup any 
# trash files that were left from earlier calls to rename, typically by 
# previous processes. The cleanup_trash method needs to be fast most of the 
# time, so the real work is only done the first time a rename operation is 
# done on an image. The image module cannot be imported when this module is
# initialized because that leads to a circular dependency.  So it is imported
# as needed.
        
trashname = "trash"
cached_image = None
def cleanup_trash(path):
        """
        Cleanup files that were left by previous calls to rename.  The path
        is used to find the image where the files were left. 
        """
        global cached_image
        if cached_image and path.startswith(cached_image.get_root()):
                return
        # Either this is the first time, or the path is in a different image
        # from what we have cached. 
        import pkg.client.image as image
        cached_image = image.Image()
        try:
                cached_image.find_root(os.path.dirname(path))
        except ValueError:
                # if path is not within an image, revert to the initial state
                cached_image = None
                return
        trashdir = os.path.join(cached_image.imgdir, trashname)
        if os.path.exists(trashdir):
                shutil.rmtree(trashdir, True)

def rename(src, dst):
        """
        Rename the src file to the dst name, deleting dst if necessary.
        """
        try:
                os.rename(src, dst)
        except OSError, err:
                if err == errno.ENOENT or not os.path.exists(dst):
                        raise
                try:
                        os.unlink(dst)
                except OSError:
                        # rename dst to trash area within image
                        import pkg.client.image as image
                        img = image.Image()
                        img.find_root(os.path.dirname(dst))
                        trashdir = os.path.join(img.imgdir, trashname)
                        if not os.path.exists(trashdir):
                                os.mkdir(trashdir)
                        fd, temp = tempfile.mkstemp(dir = trashdir)
                        os.close(fd)
                        os.unlink(temp)
                        # this rename will raise an exception if the dst file is
                        # in use and cannot be renamed.
                        os.rename(dst, temp)
                # finally rename the file
                os.rename(src, dst)
        cleanup_trash(dst)

def link(src, dst):
        copyfile(src, dst)
        
def split_path(path):
        drivepath = os.path.splitdrive(path)
        return drivepath[1].split('\\')

def get_root(path):
        drivepath = os.path.splitdrive(path)
        if drivepath[0] == "":
                return os.path.sep
        else:
                return drivepath[0] + '\\'

def copyfile(src, dst):
        shutil.copyfile(src, dst)

